#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <sys/times.h>
#include <sys/time.h>
#include <errno.h>
#include <time.h>
#include<sys/socket.h>
#include<sys/types.h>
#include<sys/stat.h>
#include<netinet/in.h>
#include<arpa/inet.h>
#include<sys/termios.h>
#include "ymbc_time.h"
#include <ssm_common.h>
#include "ssmtransport.h"
#include <pthread.h>

int sockfd,infd;
struct sockaddr_in srv;

pthread_mutex_t mutex;

/*$B%5!<%PB&%*!<%W%s(B*/
int open_tcpip(int port_num)
{
  socklen_t socklen;
  
  if((sockfd = socket(PF_INET, SOCK_STREAM, 0)) < 0){
    printf("socket err.\n");
    return 0; 
  }
  memset(&srv, 0 ,sizeof(srv));
  srv.sin_family = AF_INET;
  srv.sin_port = htons(port_num);
  
  socklen = sizeof(srv);
  if((bind(sockfd, (struct sockaddr *)&srv, socklen)) < 0){
    printf("bind");
    return 0;
  }  
  if(listen(sockfd,5) < 0){
    printf("listen");
    return 0;
  }
  /*connect*/
  puts("TCP/IP socket available");
  printf("\tport %d\n", ntohs(srv.sin_port));
  printf("\taddr %s\n",inet_ntoa(srv.sin_addr));
  if((infd = accept(sockfd, (struct sockaddr *)&srv,&socklen )) >= 0)
    puts("new connection granted.");
  return 1;  
}

int open_tcpip_cli(char* ip_address, int port_num)
{
  socklen_t socklen;
  struct sockaddr_in cli;
  
  if((infd = socket(PF_INET, SOCK_STREAM, 0)) < 0){
    printf("socket\n");
    return 0;
  }
  memset(&cli, 0 ,sizeof(cli));
  cli.sin_family = AF_INET;
  cli.sin_port = htons(port_num);
  if(!(inet_aton(ip_address,&cli.sin_addr))){
    printf("inet_aton\n");
    return 0;
  }
  socklen = sizeof(cli);
  if(connect(infd, (struct sockaddr *)&cli, socklen)){
    printf("connect error\n");
    return 0;
  }
  /*connect*/
  puts("connected to TCP/IP socket");
  printf("\tport %d\n", ntohs(cli.sin_port));
  printf("\taddr %s\n",inet_ntoa(cli.sin_addr));
  return 1;
}

/*$B?7$7$$%;%s%5$rEPO?$9$k(B*/
int regist_new_sensor(SSMT_newsensor *new_sensor, RegSensPtr sensor ,int num){
  sprintf(sensor->name, "%s",new_sensor->name); 
  if((sensor->sid = createSSM_time(new_sensor->name, new_sensor->suid, 
			       new_sensor->size, new_sensor->time,
			       new_sensor->cycle))<=0)return 0;
  sensor->size = new_sensor->size;
  sensor->property_size = new_sensor->property_size;
  sensor->data = malloc(new_sensor->size);
  if(new_sensor->property_size)
    sensor->property = malloc(new_sensor->property_size);
  else
    sensor->property = 0;
  sensor->id = num;
  if(!sensor->data)return 0;
  return 1;
}


/*TCP/IP$B$GAw?.(B*/
int send_sensor_data(RegSensPtr sensor){
  int rp,size;
  SSMT_header header;

  /*$B%X%C%@Aw?.(B*/
  header.head = SSMT_HEAD;
  header.type = SSMT_DATA;
  header.id   = sensor->id; 
  header.size = sensor->size;
  header.time = sensor->time;
  write(infd, &header, sizeof(SSMT_header));

  /*$B%G!<%?Aw?.(B*/
  rp = 0;
  while(rp < sensor->size){
    size = 1000;
    if(size > sensor->size - rp)size = sensor->size-rp;
    rp += write(infd,sensor->data + rp, size);
    //    printf(">"); 
  }
  return 1;
}

int send_newsensor(RegSensPtr sensor, char* name, int suid){
  static int id;
  SSMT_newsensor new_sensor;
  SSMT_header header;
  int num;

  sprintf(new_sensor.name, "%s",name);
  new_sensor.suid = suid;

  if((sensor->sid = openSSM(name, suid, 0)) <= 0)return 0;
  if(getSSM_info(name,suid, &new_sensor.size, &num, &new_sensor.cycle, &new_sensor.property_size)<0)return 0;
  new_sensor.time = (double)num*new_sensor.cycle;
  
  /*$B%X%C%@Aw?.(B*/
  header.head = SSMT_HEAD;
  header.type = SSMT_NEW;
  header.id   = 0; 
  header.size = sizeof(SSMT_newsensor);
  write(infd, &header, sizeof(SSMT_header));
  write(infd, &new_sensor, sizeof(SSMT_newsensor));
  /*$B%W%m%Q%F%#Aw?.(B*/
  if(new_sensor.property_size >0){
    sensor->property = malloc(new_sensor.property_size);    
    header.head = SSMT_HEAD;
    header.type = SSMT_PROPERTY;
    header.id   = id; 
    header.size = new_sensor.property_size;
    write(infd, &header, sizeof(SSMT_header));
    write(infd, &sensor->property, new_sensor.property_size);
  }else{
    sensor->property = 0;
  }
  sensor->id = id;
  sensor->suid = new_sensor.suid;
  sensor->size = new_sensor.size;
  sensor->property_size = new_sensor.property_size;
  sprintf(sensor->name,"%s",name);
  sensor->data = malloc(sensor->size);
  sensor->tid = readSSM(sensor->sid, sensor->data, &sensor->time,-1);
  
  id++;
  return 1;
}

/*$B?7$7$$%G!<%?$,$"$C$?$iAw?.$9$k(B*/
int ssm2tcp(char *filename){
  RegisterdSensor out_sensor[20];
  int out_sensor_num =0;
  int updated;
  char name[30];
  int suid,i;
  FILE *send_file;

  /*$B%;%s%5$NEPO?=hM}(B*/
  send_file = fopen(filename,"r");
  if(!send_file){
    printf("File [%s] does not exist.",filename);
    return 0;
  }

  while(fscanf(send_file, "%s %d",name,&suid) != EOF){
    if(send_newsensor(&out_sensor[out_sensor_num], name, suid)){
      printf("Sensor %s[%d] registerd.\n",
	     out_sensor[out_sensor_num].name,
	     out_sensor[out_sensor_num].suid);
      out_sensor_num++;
    }
  }
  fclose(send_file);

  /*$B%;%s%5%G!<%?$NAw?.=hM}(B*/
  while(1){
    updated = 0;
    for(i = 0;i< out_sensor_num;i++){
      if(readSSM(out_sensor[i].sid,  out_sensor[i].data, 
		 &out_sensor[i].time, out_sensor[i].tid)>0){
	out_sensor[i].tid++;
	send_sensor_data(&out_sensor[i]); /*$BAw?.$9$k(B*/      
	updated = 1;
	//	printf("s\n");
      }
    }
    if(!updated)usleep(10000);
  }
}

/*TCP/IP$B$N%a%C%;!<%8$r2r<a$7$F!"(BSSM$B$KF~NO$9$k(B*/
void tcp2ssm(void)
{
  int len;
  RegisterdSensor in_sensor[20];
  int in_sensor_num= 0;
  SSMT_header header;
  SSMT_newsensor  new_sensor;
 
  while(1){
    /*$B%X%C%@FI$_9~$_(B*/
    len = 0;
    while(len < sizeof(SSMT_header))
      len += read(infd,((char*)&header)+len,sizeof(SSMT_header)-len);
    /*header check*/
    if(header.head != SSMT_HEAD){
      printf("header_error\n");
      return;
    }
    
    /*$B$=$l$>$l$N%?%$%W$K1~$8$FFI$_9~$_(B*/
    switch(header.type){
    case SSMT_NEW:
      len = 0;
      while(len < sizeof(SSMT_newsensor))
	len += read(infd,((char*)&new_sensor)+len,sizeof(SSMT_newsensor)-len);
      if(!regist_new_sensor(&new_sensor, &in_sensor[in_sensor_num], in_sensor_num))return;
      printf("sensor %s[%d] registerd.\n",in_sensor[in_sensor_num].name,
	     in_sensor[in_sensor_num].suid);
      in_sensor_num++;
      break;
    case SSMT_PROPERTY:
      len = 0;
      while(len < header.size)
	len += read(infd,
		    ((char*)in_sensor[header.id].property)+len,
		    header.size-len);
      set_propertySSM(in_sensor[header.id].name,
		      in_sensor[header.id].suid,
		      in_sensor[header.id].property,
		      in_sensor[header.id].property_size);
      printf("set property\n");       
      break;
    case SSMT_DATA:
      len = 0;
      while(len < header.size)
	len+=read(infd,((char*)in_sensor[header.id].data)+len,header.size-len);
      writeSSM(in_sensor[header.id].sid,
	       in_sensor[header.id].data,
	       header.time);
      //printf("one data\n");
      break;
    default:
      break;
    }
  }
}

int main(int argc, char* argv[])
{
  pthread_t receive_thread;
  
  /*initialize*/
  pthread_mutex_init(&mutex,NULL);
  initSSM();
  
  if(argc > 1){
    printf("client\n");
    open_tcpip_cli(argv[1], 50000);  
  }else{
    printf("server\n");
    open_tcpip(50000);  
  }
  printf("hoge\n");
  if(pthread_create(&receive_thread,NULL,(void*)tcp2ssm, NULL) != 0){
    /*$B%(%i!<=hM}(B*/
  }


  ssm2tcp("send_sensors");

  pthread_join(receive_thread,NULL);

  return 0;
}
